@@ -68,10 +68,14 @@ Todo:
 
 """
 
-import json
-import uuid
+from __future__ import annotations
 
-from translate.lang.data import cldr_plural_categories, plural_tags
+import json
+import re
+import uuid
+from typing import BinaryIO, ClassVar, TextIO, cast
+
+from translate.lang.data import cldr_plural_categories
 from translate.misc.multistring import multistring
 from translate.storage import base
 
@@ -139,7 +143,7 @@ class BaseJsonUnit(base.DictUnit):
 class FlatUnitId(base.UnitId):
     @classmethod
     def from_string(cls, text):
-        key = text[1:] if text.startswith(".") else text
+        key = text.removeprefix(".")
         return cls([("key", key)])
 
 
@@ -150,7 +154,7 @@ class FlatJsonUnit(BaseJsonUnit):
 class JsonFile(base.DictStore):
     """A JSON file."""
 
-    UnitClass = FlatJsonUnit
+    UnitClass: ClassVar[type[BaseJsonUnit]] = FlatJsonUnit
 
     def __init__(self, inputfile=None, filter=None, **kwargs):
         """Construct a JSON file, optionally reading in from inputfile."""
@@ -166,12 +170,6 @@ class JsonFile(base.DictStore):
         if inputfile is not None:
             self.parse(inputfile)
 
-    @property
-    def plural_tags(self):
-        locale = self.gettargetlanguage()
-        locale = locale.replace("_", "-").split("-")[0] if locale else "en"
-        return plural_tags.get(locale, plural_tags["en"])
-
     def serialize(self, out):
         units = self.get_root_node()
         self.serialize_units(units)
@@ -219,27 +217,38 @@ class JsonFile(base.DictStore):
             unit.set_unitid(prev)
             yield unit
 
-    def parse(self, input):
+    def preprocess_input(self, text: str) -> str:
+        return text
+
+    def parse(self, data: str | bytes | TextIO | BinaryIO):
         """Parse the given file or file source string."""
-        if hasattr(input, "name"):
-            self.filename = input.name
+        text: str | bytes
+        if hasattr(data, "name"):
+            self.filename = data.name
         elif not getattr(self, "filename", ""):
             self.filename = ""
-        if hasattr(input, "read"):
-            src = input.read()
-            input.close()
-            input = src
-        if isinstance(input, bytes):
+        if hasattr(data, "read"):
+            # Make type checking happy
+            data = cast("BinaryIO", data)
+
+            text = data.read()
+            data.close()
+        else:
+            text = data
+
+        if isinstance(text, bytes):
             # The JSON files should be UTF-8, but implementations
             # that parse JSON texts MAY ignore the presence of a byte order mark
             # rather than treating it as an error, see RFC7159
-            input, self.encoding = self.detect_encoding(input)
-            if input is None:
+            decoded, self.encoding = self.detect_encoding(text)
+            if decoded is None:
                 raise base.ParseError(ValueError("Failed to decode JSON string."))
+            text = decoded
+        text = self.preprocess_input(text)
         try:
-            self._file = json.loads(input)
+            self._file = json.loads(text)
         except ValueError as e:
-            raise base.ParseError(e)
+            raise base.ParseError(e) from e
 
         for unit in self._extract_units(self._file, stop=self._filter):
             self.addunit(unit)
@@ -255,7 +264,7 @@ class JsonNestedFile(JsonFile):
     UnitClass = JsonNestedUnit
 
 
-class WebExtensionJsonUnit(BaseJsonUnit):
+class WebExtensionJsonUnit(FlatJsonUnit):
     def storevalues(self, output):
         value = {"message": self.target}
         if self.notes:
@@ -277,6 +286,11 @@ class WebExtensionJsonFile(JsonFile):
 
     UnitClass = WebExtensionJsonUnit
 
+    COMMENT_RE = re.compile(r'^((?:[^"\n]|"(?:[^"\\\n]|\\.)*")*?)//.*', re.MULTILINE)
+
+    def preprocess_input(self, text: str) -> str:
+        return self.COMMENT_RE.sub(r"\1", text)
+
     def _extract_units(
         self,
         data,
@@ -287,13 +301,17 @@ class WebExtensionJsonFile(JsonFile):
         last_node=None,
     ):
         for item, value in data.items():
+            if isinstance(value, str):
+                raise base.ParseError(
+                    ValueError("File is not a valid WebExtension JSON file!")
+                )
             unit = self.UnitClass(
                 value.get("message", ""),
                 item,
                 value.get("description", ""),
                 value.get("placeholders", None),
             )
-            unit.setid(item)
+            unit.setid(item, unitid=self.UnitClass.IdClass.from_key(item))
             yield unit
 
 
@@ -349,9 +367,9 @@ class I18NextUnit(JsonNestedUnit):
         if not isinstance(self.target, multistring):
             super().storevalues(output)
         else:
-            if len(self.target.strings) > len(self._store.plural_tags):
-                self.target.strings = self.target.strings[
-                    : len(self._store.plural_tags)
+            if len(self.target.strings) > len(self._store.get_plural_tags()):
+                self.target.extra_strings = self.target.extra_strings[
+                    : len(self._store.get_plural_tags()) - 1
                 ]
             self._fixup_item()
             for i, value in enumerate(self.target.strings):
@@ -394,7 +412,7 @@ class I18NextFile(JsonNestedFile):
                 plurals = []
                 plural_base = ""
                 if k in plurals_simple or k + "_plural" in plurals_simple:
-                    plural_base = k[:-7] if k.endswith("_plural") else k
+                    plural_base = k.removesuffix("_plural")
                     plurals_simple.remove(plural_base)
                     plurals = [k, k + "_plural"]
                 elif "_" in k:
@@ -439,7 +457,7 @@ class I18NextV4Unit(I18NextUnit):
 
     def _get_plural_labels(self, count):
         base_name = self._get_base_name()
-        return [f"{base_name}_{self._store.plural_tags[i]}" for i in range(count)]
+        return [f"{base_name}_{self._store.get_plural_tags()[i]}" for i in range(count)]
 
 
 class I18NextV4File(JsonNestedFile):
@@ -478,7 +496,9 @@ class I18NextV4File(JsonNestedFile):
                     plural_base, suffix = k.rsplit("_", 1)
 
                 if suffix in cldr_plural_categories:
-                    plurals = [f"{plural_base}_{suffix}" for suffix in self.plural_tags]
+                    plurals = [
+                        f"{plural_base}_{suffix}" for suffix in self.get_plural_tags()
+                    ]
 
                 if plurals:
                     sources = []
@@ -560,7 +580,7 @@ class GoTextJsonUnit(BaseJsonUnit):
     def getvalue(self):
         target = self.target
         if isinstance(target, multistring):
-            strings = self.sync_plural_count(target, self._store.plural_tags)
+            strings = self.sync_plural_count(target, self._store.get_plural_tags())
             target = {
                 "select": {
                     "feature": "plural",
@@ -570,7 +590,7 @@ class GoTextJsonUnit(BaseJsonUnit):
                 target["select"]["arg"] = self.placeholders[0]["id"]
             target["select"]["cases"] = {
                 plural: {"msg": strings[offset]}
-                for offset, plural in enumerate(self._store.plural_tags)
+                for offset, plural in enumerate(self._store.get_plural_tags())
             }
         value = {"id": self._unitid.parts if self._unitid else self.getid()}
         if self.message:
@@ -666,16 +686,16 @@ class GoTextJsonFile(JsonFile):
         out.write(b"\n")
 
 
-class GoI18NJsonUnit(BaseJsonUnit):
+class GoI18NJsonUnit(FlatJsonUnit):
     ID_FORMAT = "{}"
 
     def getvalue(self):
         target = self.target
         if isinstance(target, multistring):
-            strings = self.sync_plural_count(target, self._store.plural_tags)
+            strings = self.sync_plural_count(target, self._store.get_plural_tags())
             target = {
                 plural: strings[offset]
-                for offset, plural in enumerate(self._store.plural_tags)
+                for offset, plural in enumerate(self._store.get_plural_tags())
             }
         value = {"id": self.getid()}
         if self.notes:
@@ -730,7 +750,8 @@ class GoI18NJsonFile(JsonFile):
                 value.get("id", ""),
                 value.get("description", ""),
             )
-            unit.setid(value.get("id", ""))
+            item = value.get("id", "")
+            unit.setid(item, unitid=self.UnitClass.IdClass.from_key(item))
             yield unit
 
     def serialize(self, out):
@@ -739,7 +760,7 @@ class GoI18NJsonFile(JsonFile):
         out.write(b"\n")
 
 
-class GoI18NV2JsonUnit(BaseJsonUnit):
+class GoI18NV2JsonUnit(FlatJsonUnit):
     ID_FORMAT = "{}"
 
     def converttarget(self):
@@ -753,8 +774,8 @@ class GoI18NV2JsonUnit(BaseJsonUnit):
         if self.notes:
             target["description"] = self.notes
 
-        strings = self.sync_plural_count(self.target, self._store.plural_tags)
-        for offset, plural in enumerate(self._store.plural_tags):
+        strings = self.sync_plural_count(self.target, self._store.get_plural_tags())
+        for offset, plural in enumerate(self._store.get_plural_tags()):
             target[plural] = strings[offset]
 
         return target
@@ -802,11 +823,11 @@ class GoI18NV2JsonFile(JsonFile):
                     id,
                     value.get("description", ""),
                 )
-            unit.setid(id)
+            unit.setid(id, unitid=self.UnitClass.IdClass.from_key(id))
             yield unit
 
 
-class ARBJsonUnit(BaseJsonUnit):
+class ARBJsonUnit(FlatJsonUnit):
     ID_FORMAT = "{}"
 
     def __init__(
@@ -875,6 +896,10 @@ class ARBJsonFile(JsonFile):
         for item, value in data.items():
             if item.startswith("@"):
                 continue
+            if not isinstance(value, (str, int)):
+                raise base.ParseError(
+                    ValueError(f"Key {item!r} does not contain string: {value!r}")
+                )
             metadata = data.get(f"@{item}", {})
             unit = self.UnitClass(
                 value,
@@ -883,11 +908,11 @@ class ARBJsonFile(JsonFile):
                 metadata.get("placeholders", None),
                 metadata=metadata,
             )
-            unit.setid(item)
+            unit.setid(item, unitid=self.UnitClass.IdClass.from_key(item))
             yield unit
 
 
-class FormatJSJsonUnit(BaseJsonUnit):
+class FormatJSJsonUnit(FlatJsonUnit):
     def storevalues(self, output):
         value = {"defaultMessage": self.target}
         if self.notes:
@@ -901,7 +926,7 @@ class FormatJSJsonFile(JsonFile):
 
     See following URLs for doc:
 
-    https://formatjs.io/docs/getting-started/message-extraction/
+    https://formatjs.github.io/docs/getting-started/message-extraction/
     """
 
     UnitClass = FormatJSJsonUnit
@@ -916,10 +941,16 @@ class FormatJSJsonFile(JsonFile):
         last_node=None,
     ):
         for item, value in data.items():
+            if not isinstance(value, dict):
+                raise base.ParseError(
+                    ValueError(
+                        f"Not a valid FormatJS file should be a object: {value!r}"
+                    )
+                )
             unit = self.UnitClass(
                 value.get("defaultMessage", ""),
                 item,
                 value.get("description", ""),
             )
-            unit.setid(item)
+            unit.setid(item, unitid=self.UnitClass.IdClass.from_key(item))
             yield unit
